@@ -28,6 +28,12 @@
 #include "soup-ssl.h"
 #include "soup-uri.h"
 
+typedef enum {
+	SOUP_CONNECTION_MODE_DIRECT,
+	SOUP_CONNECTION_MODE_PROXY,
+	SOUP_CONNECTION_MODE_TUNNEL
+} SoupConnectionMode;
+
 typedef struct {
 	SoupSocket  *socket;
 
@@ -40,11 +46,15 @@ typedef struct {
 	SoupUri     *proxy_uri, *origin_uri, *conn_uri;
 	gpointer     ssl_creds;
 
+	SoupConnectionMode  mode;
+
 	SoupMessageFilter *filter;
+	GMainContext      *async_context;
 
 	SoupMessage *cur_req;
 	time_t       last_used;
 	gboolean     connected, in_use;
+	guint        timeout;
 } SoupConnectionPrivate;
 #define SOUP_CONNECTION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_CONNECTION, SoupConnectionPrivate))
 
@@ -67,6 +77,8 @@ enum {
 	PROP_PROXY_URI,
 	PROP_SSL_CREDS,
 	PROP_MESSAGE_FILTER,
+	PROP_ASYNC_CONTEXT,
+	PROP_TIMEOUT,
 
 	LAST_PROP
 };
@@ -82,6 +94,8 @@ static void clear_current_request (SoupConnection *conn);
 static void
 soup_connection_init (SoupConnection *conn)
 {
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
+	priv->timeout = 0;
 }
 
 static void
@@ -96,6 +110,8 @@ finalize (GObject *object)
 
 	if (priv->filter)
 		g_object_unref (priv->filter);
+	if (priv->async_context)
+		g_main_context_unref (priv->async_context);
 
 	G_OBJECT_CLASS (soup_connection_parent_class)->finalize (object);
 }
@@ -244,6 +260,19 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 				      "Message filter",
 				      "Message filter object for this connection",
 				      G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_ASYNC_CONTEXT,
+		g_param_spec_pointer (SOUP_CONNECTION_ASYNC_CONTEXT,
+				      "Async GMainContext",
+				      "GMainContext to dispatch this connection's async I/O in",
+				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (
+		object_class, PROP_TIMEOUT,
+		g_param_spec_uint (SOUP_CONNECTION_TIMEOUT,
+				   "Timeout value",
+				   "Value in seconds to timeout a blocking I/O",
+				   0, G_MAXUINT, 0,
+				   G_PARAM_READWRITE));
 }
 
 
@@ -289,22 +318,43 @@ set_property (GObject *object, guint prop_id,
 	case PROP_ORIGIN_URI:
 		pval = g_value_get_pointer (value);
 		priv->origin_uri = pval ? soup_uri_copy (pval) : NULL;
-		if (!priv->proxy_uri)
-			priv->conn_uri = priv->origin_uri;
-		break;
+		goto changed_uri;
+
 	case PROP_PROXY_URI:
 		pval = g_value_get_pointer (value);
 		priv->proxy_uri = pval ? soup_uri_copy (pval) : NULL;
-		if (priv->proxy_uri)
+
+	changed_uri:
+		if (priv->proxy_uri) {
 			priv->conn_uri = priv->proxy_uri;
-		else
+			if (priv->origin_uri &&
+			    priv->origin_uri->protocol == SOUP_PROTOCOL_HTTPS)
+				priv->mode = SOUP_CONNECTION_MODE_TUNNEL;
+			else
+				priv->mode = SOUP_CONNECTION_MODE_PROXY;
+		} else {
 			priv->conn_uri = priv->origin_uri;
+			priv->mode = SOUP_CONNECTION_MODE_DIRECT;
+		}
 		break;
+
 	case PROP_SSL_CREDS:
 		priv->ssl_creds = g_value_get_pointer (value);
 		break;
 	case PROP_MESSAGE_FILTER:
-		priv->filter = g_object_ref (g_value_get_pointer (value));
+		if (priv->filter)
+			g_object_unref (priv->filter);
+		priv->filter = g_value_get_pointer (value);
+		if (priv->filter)
+			g_object_ref (priv->filter);
+		break;
+	case PROP_ASYNC_CONTEXT:
+		priv->async_context = g_value_get_pointer (value);
+		if (priv->async_context)
+			g_main_context_ref (priv->async_context);
+		break;
+	case PROP_TIMEOUT:
+		priv->timeout = g_value_get_uint (value);
 		break;
 	default:
 		break;
@@ -331,7 +381,13 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_pointer (value, priv->ssl_creds);
 		break;
 	case PROP_MESSAGE_FILTER:
-		g_value_set_pointer (value, g_object_ref (priv->filter));
+		g_value_set_pointer (value, priv->filter ? g_object_ref (priv->filter) : NULL);
+		break;
+	case PROP_ASYNC_CONTEXT:
+		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
+		break;
+	case PROP_TIMEOUT:
+		g_value_set_uint (value, priv->timeout);
 		break;
 	default:
 		break;
@@ -346,7 +402,7 @@ set_current_request (SoupConnectionPrivate *priv, SoupMessage *req)
 	req->status = SOUP_MESSAGE_STATUS_RUNNING;
 	priv->cur_req = req;
 	priv->in_use = TRUE;
-	g_object_add_weak_pointer (G_OBJECT (req), (gpointer *)&priv->cur_req);
+	g_object_add_weak_pointer (G_OBJECT (req), (gpointer)&priv->cur_req);
 }
 
 static void
@@ -355,14 +411,18 @@ clear_current_request (SoupConnection *conn)
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
 	if (priv->cur_req) {
-		if (!soup_message_is_keepalive (priv->cur_req))
-			soup_connection_disconnect (conn);
-		else
-			priv->last_used = time (NULL);
+		SoupMessage *cur_req = priv->cur_req;
 
 		g_object_remove_weak_pointer (G_OBJECT (priv->cur_req),
-					      (gpointer *)&priv->cur_req);
+					      (gpointer)&priv->cur_req);
 		priv->cur_req = NULL;
+
+		if (!soup_message_is_keepalive (cur_req))
+			soup_connection_disconnect (conn);
+		else {
+			priv->last_used = time (NULL);
+			soup_message_io_stop (cur_req);
+		}
 	}
 	priv->in_use = FALSE;
 }
@@ -459,10 +519,7 @@ socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 		}
 	}
 
-	/* See if we need to tunnel */
-	if (priv->proxy_uri &&
-	    priv->origin_uri &&
-	    priv->origin_uri->protocol == SOUP_PROTOCOL_HTTPS) {
+	if (priv->mode == SOUP_CONNECTION_MODE_TUNNEL) {
 		SoupMessage *connect_msg;
 
 		connect_msg = soup_message_new_from_uri (SOUP_METHOD_CONNECT,
@@ -509,10 +566,13 @@ soup_connection_connect_async (SoupConnection *conn,
 	}
 
 	priv->socket =
-		soup_socket_client_new_async (priv->conn_uri->host,
-					      priv->conn_uri->port,
-					      priv->ssl_creds,
-					      socket_connect_result, conn);
+		soup_socket_new (SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
+				 SOUP_SOCKET_ASYNC_CONTEXT, priv->async_context,
+				 NULL);
+	soup_socket_connect (priv->socket, soup_address_new (priv->conn_uri->host,
+							     priv->conn_uri->port));
+	soup_signal_connect_once (priv->socket, "connect_result",
+				  G_CALLBACK (socket_connect_result), conn);
 	g_signal_connect (priv->socket, "disconnected",
 			  G_CALLBACK (socket_disconnected), conn);
 }
@@ -529,6 +589,7 @@ guint
 soup_connection_connect_sync (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv;
+	SoupAddress* addr;
 	guint status;
 
 	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), SOUP_STATUS_MALFORMED);
@@ -536,14 +597,21 @@ soup_connection_connect_sync (SoupConnection *conn)
 	g_return_val_if_fail (priv->socket == NULL, SOUP_STATUS_MALFORMED);
 
 	priv->socket =
-		soup_socket_client_new_sync (priv->conn_uri->host,
-					     priv->conn_uri->port,
-					     priv->ssl_creds,
-					     &status);
+		soup_socket_new (SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
+				 SOUP_SOCKET_FLAG_NONBLOCKING, FALSE,
+				 SOUP_SOCKET_TIMEOUT, priv->timeout,
+				 NULL);
+
+	addr = soup_address_new (priv->conn_uri->host,
+				 priv->conn_uri->port);
+
+	status = soup_socket_connect (priv->socket, addr);
+	g_object_unref (addr);
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 		goto fail;
 
+		
 	g_signal_connect (priv->socket, "disconnected",
 			  G_CALLBACK (socket_disconnected), conn);
 
@@ -554,10 +622,7 @@ soup_connection_connect_sync (SoupConnection *conn)
 		}
 	}
 
-	/* See if we need to tunnel */
-	if (priv->proxy_uri &&
-	    priv->origin_uri &&
-	    priv->origin_uri->protocol == SOUP_PROTOCOL_HTTPS) {
+	if (priv->mode == SOUP_CONNECTION_MODE_TUNNEL) {
 		SoupMessage *connect_msg;
 
 		connect_msg = soup_message_new_from_uri (SOUP_METHOD_CONNECT,
@@ -576,6 +641,12 @@ soup_connection_connect_sync (SoupConnection *conn)
 		}
 
 		g_object_unref (connect_msg);
+
+		if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
+			if (!soup_socket_start_proxy_ssl (priv->socket,
+							 priv->origin_uri->host))
+				status = SOUP_STATUS_SSL_FAILED;
+		}
 	}
 
 	if (SOUP_STATUS_IS_SUCCESSFUL (status))
@@ -713,7 +784,7 @@ send_request (SoupConnection *conn, SoupMessage *req)
 	}
 
 	soup_message_send_request_internal (req, priv->socket, conn,
-					    priv->proxy_uri != NULL);
+					    priv->mode == SOUP_CONNECTION_MODE_PROXY);
 }
 
 /**
