@@ -24,6 +24,7 @@ parse_response_headers (SoupMessage *req,
 			guint *content_len,
 			gpointer user_data)
 {
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (req);
 	const char *length, *enc;
 	SoupHttpVersion version;
 	GHashTable *resp_hdrs;
@@ -36,8 +37,8 @@ parse_response_headers (SoupMessage *req,
 					  (char **) &req->reason_phrase))
 		return SOUP_STATUS_MALFORMED;
 
-	if (version < req->priv->http_version)
-		req->priv->http_version = version;
+	if (version < priv->http_version)
+		priv->http_version = version;
 
 	meth_id   = soup_method_get_id (req->method);
 	resp_hdrs = req->response_headers;
@@ -68,7 +69,7 @@ parse_response_headers (SoupMessage *req,
 	 */
 	enc = soup_message_get_header (resp_hdrs, "Transfer-Encoding");
 	if (enc) {
-		if (g_strcasecmp (enc, "chunked") == 0) {
+		if (g_ascii_strcasecmp (enc, "chunked") == 0) {
 			*encoding = SOUP_TRANSFER_CHUNKED;
 			return SOUP_STATUS_OK;
 		} else
@@ -107,6 +108,7 @@ get_request_headers (SoupMessage *req, GString *header,
 		     SoupTransferEncoding *encoding,
 		     gpointer user_data)
 {
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (req);
 	gboolean proxy = GPOINTER_TO_UINT (user_data);
 	const SoupUri *uri = soup_message_get_uri (req);
 	const char *expect;
@@ -122,7 +124,7 @@ get_request_headers (SoupMessage *req, GString *header,
 		uri_string = soup_uri_to_string (uri, !proxy);
 	}
 
-	if (req->priv->http_version == SOUP_HTTP_1_0) {
+	if (priv->http_version == SOUP_HTTP_1_0) {
 		g_string_append_printf (header, "%s %s HTTP/1.0\r\n",
 					req->method, uri_string);
 	} else {
@@ -154,15 +156,33 @@ get_request_headers (SoupMessage *req, GString *header,
 
 	expect = soup_message_get_header (req->request_headers, "Expect");
 	if (expect && !strcmp (expect, "100-continue"))
-		req->priv->msg_flags |= SOUP_MESSAGE_EXPECT_CONTINUE;
+		priv->msg_flags |= SOUP_MESSAGE_EXPECT_CONTINUE;
 }
 
+/**
+ * soup_message_send_request:
+ * @req: a #SoupMessage
+ * @sock: the #SoupSocket to send @req on
+ * @is_via_proxy: %TRUE if @sock is a connection to a proxy server
+ * rather than a direct connection to the desired HTTP server
+ *
+ * Begins the process of sending @msg across @sock. (If @sock is
+ * synchronous, then soup_message_send_request() won't return until
+ * the response has been received.)
+ **/
 void
 soup_message_send_request (SoupMessage *req, SoupSocket *sock,
 			   gboolean is_via_proxy)
+{
+	soup_message_send_request_internal (req, sock, NULL, is_via_proxy);
+}
+
+void
+soup_message_send_request_internal (SoupMessage *req, SoupSocket *sock,
+				    SoupConnection *conn, gboolean is_via_proxy)
 {
 	soup_message_cleanup_response (req);
-	soup_message_io_client (req, sock,
+	soup_message_io_client (req, sock, conn,
 				get_request_headers,
 				parse_response_headers,
 				GUINT_TO_POINTER (is_via_proxy));
